<?php

namespace Drp\IOEngage;

use Carbon\Carbon;
use Drp\IOEngage\Exceptions\AuthenticationFailed;
use Drp\IOEngage\Exceptions\DelegateRetrievalFailed;
use Drp\IOEngage\Exceptions\MeetingUpdateFailed;
use Drp\IOEngage\Exceptions\SearchFailed;
use Drp\IOEngage\Responses\DelegateResponse;
use Drp\IOEngage\Responses\MeetingResponse;
use Drp\IOEngage\Responses\SearchResponse;
use Drp\IOEngage\Responses\TokenResponse;
use Illuminate\Http\Client\Response;
use Illuminate\Support\Arr;
use Illuminate\Support\Facades\Http;

class IOEngage
{
    /**
     * @var string
     */
    protected $exception;

    /**
     * Must supply Country when performing search. Default will be GB
     *
     * @param array $args
     * @throws BindingResolutionException
     * @throws Exception
     * @throws RuntimeException
     * @throws Throwable
     */
    public function search(array $args): SearchResponse
    {
        throw_unless(array_key_exists('Country', $args), SearchFailed::class);

        $response = $this->makeRequest(
            $this->buildUrlParams(
                config('ioengage.endpoints.search'),
                $args
            ),
            SearchFailed::class
        );

        return new SearchResponse($this->getContents($response));
    }

    /**
     * Can get a maximum of 500 delegates per request as defined by NGINX max URI length
     *
     * @throws BindingResolutionException
     * @throws Exception
     * @throws RuntimeException
     * @throws Throwable
     */
    public function getDelegates(array $oneKeyIds, string $country = 'GB'): DelegateResponse
    {
        $uri = $this->buildUrlParams(
            config('ioengage.endpoints.delegate') . '/' . implode(',', $oneKeyIds),
            ['Country' => $country]
        );

        $response = $this->makeRequest(
            $uri,
            DelegateRetrievalFailed::class,
        );

        return new DelegateResponse($this->getContents($response));
    }

    public function meetingUpdate(array $payload): MeetingResponse
    {
        $response = $this->makeRequest(
            config('ioengage.endpoints.meeting'),
            MeetingUpdateFailed::class,
            'post',
            [
                "Meeting" => $payload,
            ]
        );

        return new MeetingResponse($this->getContents($response, 201));
    }

    public function authenticate(): TokenResponse
    {
        if ($this->needsToAuthenticate() === false) {
            return new TokenResponse(cache('ioengage'));
        }

        $this->exception = AuthenticationFailed::class;
        try {
            $response = Http::withHeaders(config('ioengage.authentication'))
                ->post(config('ioengage.endpoints.token'));
        } catch (\Throwable $th) {
            throw new $this->exception();
        }

        $token = new TokenResponse($this->getContents($response));

        cache(['ioengage' => $token->forCache()]);

        return $token;
    }

    public function makeRequest(string $uri, string $exception, string $method = 'get', ?array $body = null): Response
    {
        $this->authenticate();

        $this->exception = $exception;

        ['access_token' => $token] = cache('ioengage');

        return Http::withHeaders([
            'Authorization' => "Bearer {$token}",
            'X-IP-ADDRESS' => config('ioengage.authentication.ip'),
        ])->$method($uri, $body);
    }

    private function getContents(Response $response, int $status = 200): array
    {
        throw_unless(
            $response->status() === $status,
            new $this->exception,
            ['message' => 'Status Code Invalid']
        );

        return json_decode($response->getBody(), true);
    }

    /**
     * Checks the current cache for the existence of credentials, then checks the expiry to determine whether the token is still valid.
     */
    private function needsToAuthenticate(): bool
    {
        $cache = cache('ioengage');
        if ($cache !== null
            && Arr::get($cache, 'access_token') !== null
            && Arr::get($cache, 'expires_in') > Carbon::now()
        ) {
            return false;
        }

        return true;
    }

    private function buildUrlParams(string $baseUri, array $args): string
    {
        return "{$baseUri}?" . implode('&', collect($args)->map(function ($value, $key): string {
            return "{$key}={$value}";
        })->toArray());
    }
}
