<?php

namespace Drp\IOEngage\Tests\Unit;

use Carbon\Carbon;
use Drp\IOEngage\Exceptions\AuthenticationFailed;
use Drp\IOEngage\Exceptions\SearchFailed;
use Drp\IOEngage\IOEngage;
use Drp\IOEngage\Tests\TestCase;
use Illuminate\Support\Arr;
use Illuminate\Support\Facades\Http;
use Spatie\Snapshots\MatchesSnapshots;

class IOEngageTest extends TestCase
{
    use MatchesSnapshots;

    protected $testDelegate = 'WUKA00000161';

    protected $testDelegates = [];

    /** @test */
    public function can_authenticate(): void
    {
        cache(['ioengage' => null]);

        $this->goodAuthPayload();
        $this->assertEmpty(cache('ioengage'));
        $this->client()->authenticate();

        $cache = cache('ioengage');

        $this->assertNotEmpty($cache);
        $this->assertArrayHasKey('access_token', $cache);
        $this->assertArrayHasKey('expires_in', $cache);
    }

    /** @test */
    public function does_not_re_authenticate_if_token_not_expired(): void
    {
        $this->mockAuth();
        $this->goodAuthPayload();
        $this->client()->authenticate();

        $cache = cache('ioengage');
        $this->assertEquals('ACCESS_TOKEN', Arr::get($cache, 'access_token'));
    }

    /** @test */
    public function throws_exception_if_authentication_payload_is_bad(): void
    {
        $this->badAuthPayload();
        $this->expectException(AuthenticationFailed::class);

        $this->client()->authenticate();
    }

    /** @test */
    public function can_get_a_single_delegate_by_id(): void
    {
        $this->mockAuth();
        $this->delegatePayload();
        $result = $this->client()->getDelegates([$this->testDelegate]);
        $this->assertMatchesJsonSnapshot($result);
    }

    /** @test */
    public function can_get_multiple_delegates_by_id_array(): void
    {
        $this->mockAuth();
        $result = $this->client()->getDelegates(['WUKA00005610', 'WUKA00005611']);
        $this->assertMatchesJsonSnapshot($result);
    }

    /** @test */
    public function can_perform_search_with_parameters(): void
    {
        $this->mockAuth();
        $result = $this->client()->search([
            'Country' => 'GB',
            'FirstName' => 'Chris',
            'City' => 'London',
        ]);
    }

    /** @test */
    public function cannot_perform_search_without_country_parameter(): void
    {
        $this->mockAuth();
        $this->expectException(SearchFailed::class);

        $this->client()->search([
            'FirstName' => 'Chris',
            'City' => 'London',
        ]);
    }

    private function client(): IOEngage
    {
        return app(IOEngage::class);
    }

    private function mockAuth(): void
    {
        cache(['ioengage' => [
            'access_token' => 'ACCESS_TOKEN',
            'expires_in' => Carbon::now()->addSeconds(600),
        ]]);
    }

    private function goodAuthPayload(): void
    {
        Http::fake([
            config('ioengage.endpoints.token') => Http::response([
                'access_token' => 'A_VERY_REAL_TOKEN',
                'token_type' => 'Bearer',
                'expires_in' => 36400,
            ], 200),
        ]);
    }

    private function badAuthPayload(): void
    {
        Http::fake([
            config('ioengage.endpoints.token') => Http::response([
                'token_type' => 'Bearer',
                'expires_in' => 36400,
            ], 200),
        ]);
    }

    private function delegatePayload(): void
    {
        $uri = config('ioengage.endpoints.delegate') . "/{$this->testDelegate}?Country=GB";
        Http::fake([
            $uri => Http::response([
                'ValidRecordCount' => 1,
                'ValidResults' => [
                    [
                        'OneKeyID' => 'WUKA00000161',
                        'Salutation' => 'Dr',
                        'Gender' => 'Female',
                        'FirstName' => 'Kevleen',
                        'LastName' => 'Jackson',
                        'Name' => 'Kevleen Jackson',
                        'Phone' => '012345678890',
                        'Email' => 'thisguy@example.com',
                        'JobTitle' => 'Dr',
                        'Company' => 'The Nutt House',
                        'CanbeMarketedTo' => true,
                        'Address' => [
                            'Line1' => '123 Fake Stret',
                            'Line2' => null,
                            'Line3' => null,
                            'City' => 'London',
                            'Postcode' => 'E69 420',
                            'County' => 'Greater London',
                            'Country' => 'GB',
                        ],
                        'Region' => 'WGB.UG2.EZ08',
                        'KeyContact' => 'false',
                    ],
                ],
                'InvalidRecordCount' => 0,
                'InvalidResults' => [],
            ], 200),
        ]);
    }

    public function setUp(): void
    {
        parent::setUp();

        config(['ioengage' => [
            'authentication' => [
                'client_id' => 'CLIENT_ID',
                'client_secret' => 'CLIENT_SECRET',
                'grant_type' => 'CLIENT_CREDENTIALS',
            ],
            'ip' => '2.3.2.1',
            'endpoints' => [
                'token' => 'https://example.com/token',
                'delegate' => 'https://example.com/delegate',
                'search' => 'https://example.com/search',
                'meeting' => 'https://example.com/meeting',
            ],
        ]]);
    }
}
